# Running programs that use DDL

Programs using DDL are run with `ddlrun`. For information on ddlrun, see:

      `$CONDA_PREFIX/doc/ddl/README.md`

# DDL functions and semantics

## Init function

```python
.Input("num_gpu: int32") => number of gpus to use on each host machine
.Output("rank: int32")   => MPI rank
.Output("size: int32")   => MPI size
.Output("gpuid: int32")  => assigned gpuid
.Attr("mode: string")    => mode, refer to $CONDA_PREFIX/doc/ddl/README.md
```

This must be called before any native TensorFlow operators. Typically,
we can execute this op on CPU using an additional session. The input is
the DDL configuration described in `$CONDA_PREFIX/doc/ddl/README.md`. This
will inform the targeted network topology and learner mapping to it. The
output consists of MPI information (rank, size) and GPU assignment.

## Bcast function

```python
.Input("input: T")   => input tensor to broadcast, always rank0 broadcasts to others
.Output("output: T") => output tensor with the broadcasted result
.Attr("T: {float}")  => supported datatype
```

`Bcast` is used to synchronize all the trainable parameters (ie. weights
and biases) before the training. `Bcast` can be called after `Init` has
been called and completed on the assign GPU device. Each and every
trainable parameter must be broadcast to ensure good convergence.

## AllReduce function

```python
.Input("input: T")          => input tensor to AllReduce
.Output("output: T")        => output tensor
.Attr("op: {'sum', 'avg'}") => reduce operation
.Attr("T: {float}")         => supported datatype
.Attr("mpi: bool = false")  => use pure MPI_Allreduce to get a baseline
.Attr("check: float = 0.0") => used to verify DDL AllReduce
```

`AllReduce` is used to synchronize the gradients of the trainable
parameters. It must be called after `Init` and `Bcast` in a typical
deep-learning network in TensorFlow. `AllReduce` is functionally
equivalent to `MPI_Allreduce` but allows DDL to exploit multi-tier
network bandwidth. The input is a tensor to `AddReduce` and the output is
the `AllReduce`d tensor.

The check attribute will:

   1. Run DDL `AllReduce`
   2. Run `MPI_Allreduce` with the same input
   3. Compare the results and flag error if the relative error it too large
      (`fabs(difference/mpi_value) > check value`) OR if "NaN" is detected.
      A dump file (`ddl.dumpXXX`) will be created if an error is
      detected.

The `check` value is tricky to set as error depends on the scale of the
values. We found 0.17 was sufficient enough to pass all the cases in
MINST with 8 learners. The reason for this error is obviously due to
different orders of additions.

`AllReduce` can be called on each trainable parameter or on concatenated
parameters (as in Caffe). If it is called on each, it is critical to
ensure the order among parameters, as also required in `MPI_Allreduce`.
If different learners `AllReduce` on different parameters, it will cause
a dead-lock situation.

## AllReduceN function

```python
.Input("input: N*T")        => input tensors to AllReduce
.Output("output: N*T")      => output tensors
.Attr("op: {'sum', 'avg'}") => reduce operation
.Attr("T: {float}")         => supported datatype
.Attr("mpi: bool = false")  => use pure MPI_Allreduce to get a baseline
.Attr("check: float = 0.0") => used to verify DDL AllReduce
```

This is a aggregated version of `AllReduce`. Essentially, this takes an
array of N tensors, performs `AllReduce` in a single shot, and return an
array of N reduced tensors. The benefit of using `AllReduceN` is better
performance and simpler integration.

&copy; Copyright IBM Corporation 2018, 2019
